// -*- C++ -*-
#include "AGILe/CCHerwig/CCHerwig.hh"
#include "ThePEG/Repository/EventGenerator.h"
#include "ThePEG/Repository/Repository.h"
#include "ThePEG/Utilities/DynamicLoader.h"
#include "ThePEG/EventRecord/Event.h"
#include "ThePEG/Vectors/HepMCConverter.h"


namespace ThePEG {
  template<> 
  struct HepMCTraits<HepMC::GenEvent> 
    : public HepMCTraitsBase<HepMC::GenEvent, HepMC::GenParticle, 
                             HepMC::GenVertex, HepMC::Polarization> {};
}
using namespace ThePEG;


namespace AGILe {

  // Standard constructor
  CCHerwig::CCHerwig() {
    // Set up particle name map
    _particleNames[ELECTRON]    = "e-";
    _particleNames[POSITRON]    = "e+";
    _particleNames[PROTON]      = "p+";
    _particleNames[ANTIPROTON]  = "pbar-";
    // NEUTRON ANTINEUTRON PHOTON MUON ANTIMUON TAU ANTITAU
    // NU_E NU_EBAR NU_MU NU_MUBAR NU_TAU NU_TAUBAR PIPLUS PIMINUS PIZERO
    // PHOTOELECTRON PHOTOPOSITRON PHOTOMUON PHOTOANTIMUON PHOTOTAU PHOTOANTITAU

    DynamicLoader::appendPath(HWLIBDIR);
    DynamicLoader::appendPath(THEPEGLIBDIR);
    Repository::load(string(HWDATADIR) + "/HerwigDefaults.rpo");
    Repository::exec("mkdir /Agile" , cout);
    Repository::exec("cd /Agile" , cout);

    // Make subprocess handler (needs to be initial-state specific)
    Repository::exec("create ThePEG::SubProcessHandler AgileSubProc" , cout);

    // Make general event handler
    Repository::exec("create ThePEG::StandardEventHandler AgileHandler", cout);
    Repository::exec("create ThePEG::LuminosityFunction AgileLuminosity", cout);
    Repository::exec("insert AgileHandler:SubProcessHandlers[0] AgileSubProc", cout);
    Repository::exec("set AgileHandler:Sampler /Herwig/ACDCSampler", cout);
    Repository::exec("set AgileHandler:CascadeHandler /Herwig/Shower/ShowerHandler", cout);
    Repository::exec("set AgileHandler:HadronizationHandler /Herwig/Hadronization/CluHadHandler", cout);
    Repository::exec("set AgileHandler:DecayHandler /Herwig/Decays/DecayHandler", cout);
    // set AgileHandler:BinStrategy 2
    // Uncomment to get QED radiation off W/Z decay products in hard process:
    // #insert AgileHandler:PostSubProcessHandlers[0] /Defaults/Handlers/QED/QEDRadiationHandler

    Repository::exec("create ThePEG::EventGenerator AgileGenerator" , cout);
    Repository::exec("set AgileGenerator:RandomNumberGenerator /Herwig/Random" , cout);
    Repository::exec("set AgileGenerator:StandardModelParameters /Herwig/Model" , cout);

    Repository::exec("create ThePEG::Cuts AgileCuts", cout);
    // Cut on the CMS energy for photon radiation
    Repository::exec("set AgileCuts:MHatMin 22.36", cout);
    // # create the cuts object for hadron collisions
    // create ThePEG::Cuts QCDCuts
    // # cuts on all process
    // set QCDCuts:ScaleMin 1.0*GeV
    // set QCDCuts:X1Min 0.00001
    // set QCDCuts:X2Min 0.00001
    // set QCDCuts:MHatMin 20.
    // # cut on jet pt 
    // create ThePEG::SimpleKTCut JetKtCut SimpleKTCut.so
    // set JetKtCut:Matcher /Defaults/Matchers/MatchStandardQCDParton
    // set JetKtCut:MinKT 20.0*GeV
    // # cut on photon
    // create ThePEG::SimpleKTCut PhotonKtCut SimpleKTCut.so
    // set PhotonKtCut:Matcher /Defaults/Matchers/MatchPhoton
    // set PhotonKtCut:MinKT 20.0*GeV
    // # cut on leptons
    // create ThePEG::SimpleKTCut LeptonKtCut SimpleKTCut.so
    // set LeptonKtCut:Matcher /Defaults/Matchers/MatchLepton
    // set LeptonKtCut:MinKT 0.0*GeV
    // #cut on top quarks
    // create ThePEG::SimpleKTCut TopKtCut SimpleKTCut.so
    // set TopKtCut:Matcher /Defaults/Matchers/MatchTop
    // set TopKtCut:MinKT 0.0*GeV
    // # create a cut on the invariant mass of lepton pairs
    // create ThePEG::V2LeptonsCut MassCut V2LeptonsCut.so
    // set MassCut:MaxM 14000*GeV
    // set MassCut:Families All
    // set MassCut:CComb All
    // set MassCut:MinM 20.*GeV
    // set MassCut:MaxM 14000.*GeV
    // # insert into hadron cuts
    // insert QCDCuts:OneCuts[0] JetKtCut
    // insert QCDCuts:OneCuts[1] PhotonKtCut
    // insert QCDCuts:OneCuts[2] LeptonKtCut
    // insert QCDCuts:OneCuts[3] TopKtCut
    // insert QCDCuts:MultiCuts[0] MassCut
    
    Repository::exec("set AgileGenerator:RandomNumberGenerator:Seed 31122001", cout);
    Repository::exec("set AgileGenerator:DebugLevel 1", cout);
    Repository::exec("set AgileGenerator:PrintEvent 0", cout);
    Repository::exec("set AgileGenerator:MaxErrors 10000", cout);
    Repository::exec("set AgileGenerator:NumberOfEvents 1000000000", cout);
  }


  // Set up initial state from supplied params
  Status CCHerwig::setGenSpecificInitialState(PdgCode p1, double e1, 
                                              PdgCode p2, double e2) {
    cout << "Setting initial state..." << endl;

    // Make string versions of the energies
    stringstream e1_ss, e2_ss;
    e1_ss << e1;
    e2_ss << e2;

    Repository::exec("set AgileLuminosity:BeamEmaxA " + e1_ss.str(), cout);
    Repository::exec("set AgileLuminosity:BeamEmaxB " + e2_ss.str(), cout);
    Repository::exec("set AgileHandler:BeamA /Herwig/Particles/" + _particleNames[p1], cout);
    Repository::exec("set AgileHandler:BeamB /Herwig/Particles/" + _particleNames[p2], cout);

    // e+ e- default matrix element
    if ((p1 == ELECTRON || p1 == POSITRON) && (p2 == POSITRON || p2 == ELECTRON)) {
      Repository::exec("cp /Herwig/Partons/EEExtractor AgileExtractor", cout);
      Repository::exec("insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEee2gZ2qq" , cout);
    }

    // hadron-hadron default matrix element
    if ((p1 == PROTON || p1 == ANTIPROTON) || (p2 == PROTON || p2 == ANTIPROTON)) {
      Repository::exec("cp /Herwig/Partons/QCDExtractor AgileExtractor", cout);
      Repository::exec("insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEqq2gZ2ff", cout);
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEqq2W2ff
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEGammaGamma
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEGammaJet
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEHiggs
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEHiggsJet
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEQCD2to2
      // #insert AgileSubProc:MatrixElements[0] /Herwig/MatrixElements/MEHeavyQuark
    }

    // Set parton extractor
    Repository::exec("set AgileSubProc:PartonExtractor AgileExtractor", cout);

    _initialstateset = true;
    return SUCCESS;
  }
  

  // Run generator initialization
  Status CCHerwig::initialize() {
    Generator::initialize();
    cout << "CCHerwig initialising..." << endl;
    
    try {
      Repository::exec("set AgileHandler:LuminosityFunction AgileLuminosity", cout);
      Repository::exec("set AgileGenerator:EventHandler AgileHandler" , cout);
      Repository::exec("set AgileGenerator:EventHandler:Cuts AgileCuts", cout);
      
      //cout << "Getting Herwig++ generator from repo..." << endl;
      EGPtr tmp = BaseRepository::GetObject<EGPtr>("/Agile/AgileGenerator");
      //cout << "Herwig++ making run..." << endl;
      eg = Repository::makeRun(tmp, "Agile");
      //cout << "Herwig++ initialising..." << endl;
      eg->initialize();
    }
    catch ( std::exception & e ) {
      cerr << e.what() << endl;
      return FAILURE;
    }

    _initialized = true;
    cout << "Herwig++ initialised" << endl;
    return SUCCESS;
  }


  Status CCHerwig::setSeed(const int seed) {
    Generator::setSeed(seed);
    setParam("AgileGenerator:RandomNumberGenerator:Seed", seed);
    return SUCCESS;
  }


  // Set parameters.
  Status CCHerwig::setParam(const string& name, const string& value) {
    Generator::setParam(name, value);
    Repository::exec("set " + name + " " + value, cout);
    return SUCCESS;
  }

  
  // Run the generator for one event.
  Status CCHerwig::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    //cout << "Making event" << endl;
    try {
      if (eg) {
        // Generate an event and convert to a HepMC::GenEvent
        EventPtr event = eg->shoot();
        HepMCConverter<HepMC::GenEvent>::convert(*event, evt, false, GeV, millimeter);
      } else {
        cerr << "eg = nil" << endl;
      }      
    }
    catch ( std::exception & e ) {
      cerr << e.what() << endl;
      return FAILURE;
    }
    //cout << "Event made" << endl;
    return SUCCESS;
  }
  

  /// Tidy up, print out run stats, etc.
  Status CCHerwig::finalize() {
    cout << "Finalising..." << endl;
    assert(eg);
    eg->finalize();
    Repository::cleanup();
    return SUCCESS;
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::CCHerwig(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
