#include "AGILe/Sherpa/Sherpa.hh"
#include <iostream>
#include <sstream>

#include "Sherpa.H"
#include "Input_Output_Handler.H"
#include "HepMC2_Interface.H"
#include "Event_Handler.H"
#include "Data_Read.H"
#include "Run_Parameter.H"
#include "Initialization_Handler.H"
#include "Read_Write_Base.H"
#include "Random.H"
using namespace SHERPA;
using namespace ATOOLS;


/// Implementation of Sherpa generator interface.
/// @author Andy Ilott
/// @author Andy Buckley
namespace AGILe {

  // Standard constructor
  Sherpa::Sherpa() {
    // Set up particle names map
    _particleNames[ELECTRON]    =    "11";
    _particleNames[POSITRON]    =   "-11";
    _particleNames[PROTON]      =  "2212";
    _particleNames[ANTIPROTON]  = "-2212";
    //   _particleNames[NEUTRON]     = "N";
    //   _particleNames[ANTINEUTRON] = "NBAR    ";
    //   _particleNames[PHOTON]      = "GAMMA   ";
    //   _particleNames[MUON]        = "MU-     ";
    //   _particleNames[ANTIMUON]    = "MU+     ";
    //   // TAU, ANTITAU
    //   _particleNames[NU_E]        = "NU_E    ";
    //   _particleNames[NU_EBAR]     = "NU_EBAR ";
    //   _particleNames[NU_MU]       = "NU_MU   ";
    //   _particleNames[NU_MUBAR]    = "NU_MUBAR";
    //   _particleNames[NU_TAU]      = "NU_TAU  ";
    //   _particleNames[NU_TAUBAR]   = "NU_TAUBR";
    //   _particleNames[PIPLUS]      = "PI+     ";
    //   _particleNames[PIMINUS]     = "PI-     ";
    //   // PIZERO
    //   // PHOTOELECTRON, PHOTOPOSITRON,
    //   // PHOTOMUON,     PHOTOANTIMUON,
    //   // PHOTOTAU,      PHOTOANTITAU,
  }


  Status Sherpa::setSeed(const int value) {
    ATOOLS::ran.SetSeed(value);
    return SUCCESS;
  }


  // Set up initial state from supplied parameters
  Status Sherpa::setGenSpecificInitialState(PdgCode p1, double e1, 
                                            PdgCode p2, double e2) {
    cout << "Setting initial state..." << endl;
    setParam("BEAM_ENERGY_1", e1);    setParam("BEAM_ENERGY_2", e2);
    setParam("BEAM_1", _particleNames[p1]);
    setParam("BEAM_2", _particleNames[p2]);
    //_initialstateset = true;
    return SUCCESS;
  }


  // Run generator initialization
  Status Sherpa::initialize() {
    Generator::initialize();
    cout << "Initialising..." << endl; 

    // Set up a dummy output so that Sherpa gives us HepMC output:
    _parameters["EVENT_MODE"] = "HepMC";

    // Tell Sherpa not to produce its own output:
    _parameters["ANALYSIS"] = "0"; 

    // Tell Sherpa where to load its integration results.
    // If the directory doesn't exist, these will not be cached between runs.
    if (_parameters.find("RESULT_DIRECTORY") == _parameters.end()) {
      _parameters["RESULT_DIRECTORY"] = "AGILeSherpaIntResults"; 
    }

    // Tell Sherpa where to load its integration results.
    // If the directory doesn't exist, these will not be cached between runs.
    if (_parameters.find("SHERPA_CPP_PATH") == _parameters.end()) {
      // Note: this path must be absolute.
      /// @todo Think about placing these somewhere specific to the model being used, cf. hash codes.
      // Relevant files which would require a clean or a different hash code are:
      //  {Processes,Particle,ME,Integration,Selector}.dat
      _parameters["SHERPA_CPP_PATH"] = string(getenv("PWD")) + "/AGILeSherpaProcessLibs";
    }

    // Specify the path which Sherpa uses to find configuration files.
    /// @todo Intention is that SHERPA will use PATH to look for a directory
    /// relative to PWD, and fall back if needed to SHERPADATADIR/Run/$PATH
    if (_parameters.find("PATH") == _parameters.end()) {
      cerr << "Currently need to specify a PATH param to SHERPA" << endl;
      throw runtime_error("Currently need to specify a PATH param to SHERPA");
      /// @todo Get this working without a prefix path: Sherpa change.
      //_parameters["PATH"] = "LHC";
      //_parameters["PATH"] = string(SHERPADATADIR) + "/Run/" + LEP91;
    }


    // Convert the parameter map into a vector which can be passed to Sherpa.
    vector<char*> c;
    c.push_back("DummyZerothParam");
    for (map<string,string>::const_iterator it = _parameters.begin(); it != _parameters.end(); ++it) {
      string command = (it->first + "=" + it->second);
      char* part = (char*) malloc((command.length()+1)*sizeof(char));
      c.push_back(strcpy(part, command.c_str()));
    }
    try {
      cout << "Initialising run:" << endl;
      _sherpa.InitializeTheRun(c.size(), &c.front());
    } catch (ATOOLS::Exception& e) {
      cerr << e.Info() << endl;
      if (e.Type() == ATOOLS::ex::normal_exit) {
        cerr << "Normal exit..." << endl;
        exit(1);
      }
      //exit(1);
      cerr << "Unexpected error during Sherpa run initialisation..." << endl;
      throw(e);
    }
    cout << "Initialising event handler:" << endl;
    _sherpa.InitializeTheEventHandler(); 

    _initialized = true;
    return SUCCESS;
  }


  /// Run the generator for one event
  Status Sherpa::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    _sherpa.GenerateOneEvent(); 
    fillEvent(evt);
    //_sherpa.SummarizeRun(); 
    return SUCCESS; 
  }


  /// Fill a HepMC event
  Status Sherpa::fillEvent(HepMC::GenEvent& evt) {
    SHERPA::Input_Output_Handler* ioh = _sherpa.GetIOHandler();
    //cout << "#### Sherpa IO Handler pointer: " << ioh << endl;
    SHERPA::HepMC2_Interface* hmc2if = ioh->GetHepMC2Interface();
    //cout << "#### Sherpa HepMC interface pointer: " << hmc2if << endl;
    HepMC::GenEvent* event = hmc2if->GenEvent();
    //cout << "#### Sherpa GenEvent pointer: " << event << endl;

    cout << "*** Original HepMC event:" << endl;
    event->print();
    /// @todo HepMC event copying problem?
    evt = *event;
    cout << "@@@ Copied HepMC event:" << endl;
    evt.print();
    clearEvent(*event);
    return SUCCESS;
  }


  /// Tidy up, print out run stats, etc.
  Status Sherpa::finalize() {
    cout << "Finalising..." << endl;
    return SUCCESS;
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::Sherpa(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
